@@ -1,14 +1,17 @@
+import multiprocessing
+import random
 import sys
 import time
 import urllib.parse
 from contextlib import contextmanager
 from dataclasses import dataclass
-from multiprocessing import Process
+from multiprocessing.context import BaseContext
 from typing import Any, Dict, List, Optional, Tuple
 
 import requests
+from requests.adapters import HTTPAdapter, Retry
 
-from .check_type import CheckTypeException, OptionalKey, check_type
+from .check_type import CheckTypeException, OptionalKey, OptionalValue, check_type
 
 
 @dataclass
@@ -28,6 +31,10 @@ class QueueCounts:
     expired: int
     completed: int
 
+    # Won't be set if a time window wasn't specified in the request, or if the
+    # server is old enough to not support rate estimation.
+    rate: Optional[float] = None
+
 
 class TasqClient:
     """
@@ -42,13 +49,16 @@ class TasqClient:
     :param username: optional username for basic authentication.
     :param password: optional password for basic authentication.
     :param max_timeout: the maximum amount of time (in seconds) to wait
-                        between attempts to pop a task in pop_running_task().
-                        Lower values mean waiting less long in the case that
-                        a new task is pushed or all tasks are finished.
+                        between attempts to pop a task in pop_running_task(),
+                        or push a task in push_blocking().
+                        Lower values mean waiting less long to pop in the case
+                        that a new task is pushed or all tasks are finished.
     :param task_timeout: if specified, override the timeout on the server with
                          a custom timeout. This can be useful if we know we
                          will be sending frequent keepalives, but the server
                          has a longer timeout period.
+    :param retry_server_errors: if True, retry requests if the server returns
+                                certain 5xx status codes.
     """
 
     def __init__(
@@ -60,6 +70,7 @@ class TasqClient:
         password: Optional[str] = None,
         max_timeout: float = 30.0,
         task_timeout: Optional[float] = None,
+        retry_server_errors: bool = True,
     ):
         self.base_url = base_url.rstrip("/")
         self.keepalive_interval = keepalive_interval
@@ -68,18 +79,74 @@ class TasqClient:
         self.password = password
         self.max_timeout = max_timeout
         self.task_timeout = task_timeout
+        self.retry_server_errors = retry_server_errors
         self.session = requests.Session()
-        if username is not None or password is not None:
-            assert username is not None and password is not None
-            self.session.auth = (username, password)
+        self._configure_session()
+        self.mp_context = multiprocessing.get_context("spawn")
 
-    def push(self, contents: str) -> str:
-        """Push a task and get its resulting ID."""
-        return self._post_form("/task/push", dict(contents=contents), type_template=str)
+    def push(self, contents: str, limit: int = 0) -> Optional[str]:
+        """
+        Push a task and get its resulting ID.
 
-    def push_batch(self, ids: List[str]) -> List[str]:
-        """Push a batch of tasks and get their resulting IDs."""
-        return self._post_json("/task/push_batch", ids, type_template=[str])
+        If limit is specified, then the task will not be pushed if the queue is
+        full, in which case None is returned.
+        """
+        return self._post_form(
+            f"/task/push", dict(contents=contents, limit=limit), type_template=OptionalValue(str)
+        )
+
+    def push_batch(self, ids: List[str], limit: int = 0) -> Optional[List[str]]:
+        """
+        Push a batch of tasks and get their resulting IDs.
+
+        If limit is specified, then tasks will not be pushed if the queue does
+        not have room for all the tasks at once, in which case None is
+        returned.
+
+        If limit is negative, then (-limit + batch_size) is used as the limit.
+        This effectively limits the size of the queue before a push rather than
+        after the push, to prevent large batches from being less likely to be
+        pushed than larger batches.
+        """
+        if limit < 0:
+            limit = -limit + len(ids)
+        return self._post_json(
+            f"/task/push_batch?limit={limit}", ids, type_template=OptionalValue([str])
+        )
+
+    def push_blocking(
+        self, contents: List[str], limit: int, init_wait_time: float = 1.0
+    ) -> List[str]:
+        """
+        Push one or more tasks atomically and block until they are pushed.
+
+        If the queue cannot fit the batch, this will wait to retry with random
+        exponential backoff. Backoff is randomized to mitigate starvation.
+
+        See push_batch() for details on passing a negative limit to avoid
+        starvation of larger batches when pushing from multiple processes.
+
+        Unlike push_batch(), the ids returned by this method will never be
+        None, since all tasks must be pushed.
+        """
+        assert isinstance(
+            contents, (list, tuple)
+        ), f"expected a list of task contents, got object of type {type(contents)}"
+        assert (
+            init_wait_time <= self.max_timeout
+        ), f"wait time {init_wait_time=} should not be larger than {self.max_timeout=}"
+        assert limit < 0 or limit >= len(contents)
+
+        cur_wait = init_wait_time
+        while True:
+            ids = self.push_batch(contents, limit=limit)
+            if ids is not None:
+                return ids
+            timeout = cur_wait * random.random()
+            time.sleep(timeout)
+            # Use summation instead of doubling to prevent really rapid
+            # growth of cur_wait with low probability.
+            cur_wait = min(cur_wait + timeout, self.max_timeout)
 
     def pop(self) -> Tuple[Optional[Task], Optional[float]]:
         """
@@ -173,7 +240,9 @@ class TasqClient:
         while True:
             task, timeout = self.pop()
             if task is not None:
-                rt = RunningTask(self, id=task.id, contents=task.contents)
+                rt = RunningTask(
+                    self, id=task.id, contents=task.contents, mp_context=self.mp_context
+                )
                 try:
                     yield rt
                     rt.completed()
@@ -186,19 +255,46 @@ class TasqClient:
                 yield None
                 return
 
-    def counts(self) -> QueueCounts:
+    def counts(self, rate_window: int = 0) -> QueueCounts:
         """Get the number of tasks in each state within the queue."""
         data = self._get(
-            "/counts",
-            dict(
-                pending=int,
-                running=int,
-                expired=int,
-                completed=int,
-            ),
+            f"/counts?window={rate_window}",
+            {
+                "pending": int,
+                "running": int,
+                "expired": int,
+                "completed": int,
+                OptionalKey("minute_rate"): float,
+            },
         )
         return QueueCounts(**data)
 
+    def __getstate__(
+        self,
+    ):
+        res = self.__dict__.copy()
+        del res["session"]
+        return res
+
+    def __setstate__(self, state: Dict[str, Any]):
+        self.__dict__ = state
+        self.session = requests.Session()
+        self._configure_session()
+
+    def _configure_session(self):
+        if self.username is not None or self.password is not None:
+            assert self.username is not None and self.password is not None
+            self.session.auth = (self.username, self.password)
+        if self.retry_server_errors:
+            retries = Retry(
+                total=10,
+                backoff_factor=1.0,
+                status_forcelist=[500, 502, 503, 504],
+                allowed_methods=False,
+            )
+            for schema in ("http://", "https://"):
+                self.session.mount(schema, HTTPAdapter(max_retries=retries))
+
     def _get(
         self, path: str, type_template: Optional[Any] = None, supports_timeout: bool = False
     ) -> Any:
@@ -229,7 +325,8 @@ class TasqClient:
         )
 
     def _url_for_path(self, path: str, supports_timeout: bool) -> str:
-        result = self.base_url + path + "?context=" + urllib.parse.quote(self.context)
+        separator = "?" if "?" not in path else "&"
+        result = self.base_url + path + separator + "context=" + urllib.parse.quote(self.context)
         if supports_timeout and self.task_timeout is not None:
             result += "&timeout=" + urllib.parse.quote(f"{self.task_timeout:f}")
         return result
@@ -242,19 +339,16 @@ class RunningTask(Task):
     cancel() or completed() is called.
     """
 
-    def __init__(self, client: TasqClient, *args, **kwargs):
+    def __init__(
+        self, client: TasqClient, *args, mp_context: Optional[BaseContext] = None, **kwargs
+    ):
         super().__init__(*args, **kwargs)
         self.client = client
-        self._proc = Process(
+        self._proc = (mp_context or multiprocessing).Process(
             target=RunningTask._keepalive_worker,
             name="tasq-keepalive-worker",
             args=(
-                client.base_url,
-                client.context,
-                client.username,
-                client.password,
-                client.task_timeout,
-                client.keepalive_interval,
+                client,
                 self.id,
             ),
             daemon=True,
@@ -274,27 +368,15 @@ class RunningTask(Task):
 
     @staticmethod
     def _keepalive_worker(
-        base_url: str,
-        context: str,
-        username: Optional[str],
-        password: Optional[str],
-        task_timeout: Optional[float],
-        interval: float,
+        client: TasqClient,
         task_id: str,
     ):
-        client = TasqClient(
-            base_url,
-            context=context,
-            username=username,
-            password=password,
-            task_timeout=task_timeout,
-        )
         while True:
             try:
                 client.keepalive(task_id)
             except Exception as exc:  # pylint: disable=broad-except
                 print(f"exception in tasq keepalive worker: {exc}", file=sys.stderr)
-            time.sleep(interval)
+            time.sleep(client.keepalive_interval)
 
 
 class TasqRemoteError(Exception):
